#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>

#include "xmem/src/include/xmem.h"
#include "include/webFactory.h"
#include "include/defaults.h"
#include "include/alterEgo.h"

#define DEF_BSIZE 8192
#define WF_QUERY_2XCRLF(b) ((*(b) == '\n') || (*(b - 1) == '\n')) && ((*(b - 2) == '\n') || (*(b - 3) == '\n'))
#define IS_DOUBLE_DOT(b) ((*(b) == '.') && (*(b + 1) == '.'))

#define WF_STATUSES_COUNT 6

struct wf_statuses_s wf_statuses[] = {
	{200, "OK", 2},
	{400, "Bad Request", 11},
	{403, "Forbidden", 9},
	{404, "Not Found", 9},
	{500, "Internal Server Error", 21},
	{503, "Service Unavailable", 19}
};

#define WF_PROTOCOLS_COUNT 3

struct wf_protocols_s wf_protocols[] = {
	{0.9, "HTTP/0.9"},
	{1.0, "HTTP/1.0"},
	{1.1, "HTTP/1.1"}
};

#define WF_METHODS_COUNT 2

struct wf_methods_s wf_methods[] = {
	{"GET"},
	{"POST"}
};

#define WF_HEADERS_COUNT 4

struct wf_headers_s wf_headers[] = {
	{"Date"},
	{"Connection"},
	{"Server"},
	{"Connection"}
};

struct factoryWorker_s *lw;

size_t wf_getFileSize(int fd) {
	size_t size, current;

	current = lseek(fd, 0L, SEEK_CUR);
	size = lseek(fd, 0L, SEEK_END);
	lseek(fd, current, SEEK_SET);

	return size;
}

/*
*	Prepare a client to assign to the worker
*/
struct wf_client_s *wf_workerClientDeclare(
	struct wf_worker_s *worker,
	int socket
) {
	struct wf_client_s *client;
	
	client = XMALLOC(struct wf_client_s, 1);
	
	client->socket = socket;
	client->bsize = DEF_BSIZE;
	client->bused = 0;
	client->stage = 0;
	client->bbtick = DEFAULT_WORKER_PUMPOVER_BYTES_BY_TICK;
	client->rawHttpQuery = XMALLOC(char, DEF_BSIZE);
	client->httpResponse = XMALLOC(char, DEF_BSIZE);
	client->parsedHttpQuery = NULL;
	client->workerOwner = worker;
	client->maxTicks = 8;
	
	if(worker->lClient == NULL) {
		worker->fClient = client;
		worker->lClient = client;
		client->prev = client;
		client->next = client;
	} else {
		client->prev = worker->lClient;
		client->next = worker->fClient;
		worker->lClient->next = client;
		worker->fClient->prev = client;
		worker->lClient = client;
	}
	
	worker->totalClients++;
	
	return client;
}

/*
*	Remove a full-handled client from the worker's stream
*/
void wf_workerClientRemove(
	struct wf_worker_s *worker,
	struct wf_client_s *client
) {
	struct wf_client_s *n, *p;

	worker->totalClients--;
	worker->cClient = NULL;

	close(client->target);
	close(client->socket);

	n = client->next;
	p = client->prev;

	if((n != client) && (p != client)) {
		n->prev = p;
		p->next = n;
	} else {
		worker->fClient = NULL;
		worker->lClient = NULL;
	}

	if(client->parsedHttpQuery) {
		mario_flushParsed(client->parsedHttpQuery);
	}
	
	XFREE(client->rawHttpQuery);
	XFREE(client->httpResponse);
	XFREE(client);
}

/*
*	Select a unhandled client from the worker's stream
*/
struct wf_client_s *wf_selectClient(
	struct wf_worker_s *worker
) {
	if(worker->cClient == NULL) {
		worker->cClient = worker->fClient;
	} else {
		worker->cClient = worker->cClient->next;
	}

	return worker->cClient;
}

/*
*	Worker's main thread which handle every assigned client
*/
void *wf_workerThread(
	void *fparam
) {
	struct wf_worker_s *worker;
	struct wf_client_s *client;
	struct wf_worker_waitClient *f;
	int newClient;
	int isNewClient;
	int z, ticks = 0;
	int targetOffset;
	int n;
	char buffer[8192];

	worker = (struct wf_worker_s *)fparam;

	while(worker->up) {
		pthread_mutex_lock(&worker->clientsList_op);
		
		if(worker->waitClients->first != NULL) {
			newClient = worker->waitClients->first->socket;

			if(worker->waitClients->first == worker->waitClients->last) {
				XFREE(worker->waitClients->first);
				worker->waitClients->first = NULL;
				worker->waitClients->last = NULL;
			} else {
				f = worker->waitClients->first;
				worker->waitClients->first = worker->waitClients->first->next;
				XFREE(f);
			}

			isNewClient = 1;
		} else {
			isNewClient = 0;
		}

		pthread_mutex_unlock(&worker->clientsList_op);
		
		if(isNewClient == 1) {
			if(newClient >= 0) {
				fcntl(newClient, F_SETFL, O_NONBLOCK);
				wf_workerClientDeclare(worker, newClient);
			}
		}
		
		client = wf_selectClient(worker);

		if(client != NULL) {
			if(client->stage == 0) {
				ticks = 0;

				while(1) {
					if(ticks > client->maxTicks) {
						break; // ticks out!
					}

					if(client->bsize - client->bused <= 0) {
						wf_resetConnection(client->socket);
					}
					
					ticks++;
					z = read(
						client->socket,
						client->rawHttpQuery + client->bused,
						client->bsize - client->bused
					);
					
					client->bused += (z > 0) ? (z) : (0);
					
					if(
						((client->bused >= 3) && WF_QUERY_2XCRLF(client->rawHttpQuery + client->bused)) || (z == 0)
					) {
						wf_clientRelease(client);
						break;
					}
				}
			}

			if(client->stage == 1) {
				client->parsedHttpQuery = mario_parse(client->rawHttpQuery);

				if(*client->parsedHttpQuery->requestLine->target == '/') {
					targetOffset = 1;
				} else {
					targetOffset = 0;		
				}

				client->target = open(
					client->parsedHttpQuery->requestLine->target + targetOffset,
					O_RDONLY | O_NONBLOCK
				);

				if(client->target > -1) {
					client->targetFound = 1;
					client->bytesGiven = 0;
					client->targetSize = wf_getFileSize(client->target);
				} else {
					client->targetFound = 0;
				}

				wf_clientRelease(client);
			}
			
			if(client->stage == 2) {
				if(client->targetFound == 1) {
					n = sprintf(
						buffer,
						"HTTP/1.1 200 OK\r\nContent-Type: text/html; charset=UTF-8\r\nContent-Length: %zu\r\nConnection: close\r\n\r\n",
						client->targetSize
					);
				} else {
					n = sprintf(
						buffer,
						"HTTP/1.1 404 Not found\r\nContent-Type: text/html; charset=UTF-8\r\nContent-Length: 3\r\nConnection: close\r\n\r\n404"
					);
				}

				write(client->socket, buffer, n);
				wf_clientRelease(client);
			}
			
			if(client->stage == 3) {
				if(client->targetFound == 1) {
					while(1) {
						if(ticks > client->maxTicks) {
							break; // ticks out!
						}
						
						ticks++;
						z = read(
							client->target,
							buffer,
							DEFAULT_WORKER_PUMPOVER_BYTES_BY_TICK
						);
						
						if(z == 0) {
							wf_clientRelease(client);
							break;
						} else {
							client->bytesGiven += z;
							write(
								client->socket,
								buffer,
								z
							);
						}
					}
				} else {
					wf_clientRelease(client);
				}
			}
			
			if(client->stage == 4) {
				wf_workerClientRemove(worker, client);
			}
		}
	}
	
	return fparam;
}

void wf_throwClient(
	int socket
) {
	struct wf_worker_s *worker_ma, *worker;
	struct wf_worker_waitClient *wc;
	int i, min;

	worker = wf.firstWorker;
	worker_ma = wf.firstWorker;
	min = worker->totalClients;
	
	for(i = 0; i < wf.workersCount; i++) {
		if(worker->totalClients < min) {
			worker_ma = worker;
		}
		
		worker = worker->next;
	}

	if(!worker_ma) {
		wf_resetConnection(socket);
	} else {
		wc = XMALLOC(struct wf_worker_waitClient, 1);
		wc->socket = socket;
		wc->next = NULL;

		pthread_mutex_lock(&worker_ma->clientsList_op);
		
		if(worker_ma->waitClients->first == NULL) {
			worker_ma->waitClients->first = wc;
			worker_ma->waitClients->last = wc;
		} else {
			worker_ma->waitClients->last->next = wc;
			worker_ma->waitClients->last = wc;
		}

		pthread_mutex_unlock(&worker_ma->clientsList_op);
	}
}

void wf_workerHotCreate(
	void
) {
	struct wf_worker_s *worker;

	// create a worker instance
	worker = XMALLOC(struct wf_worker_s, 1);
	worker->waitClients = XMALLOC(struct wf_waitClientList, 1);
	
	worker->up = 1;
	worker->totalClients = 0;
	worker->fClient = NULL;
	worker->lClient = NULL;
	worker->waitClients->first = NULL;
	worker->waitClients->last = NULL;
	worker->waitClients->total++;

	pthread_mutex_init(&worker->clientsList_op, NULL);
	pthread_create(&worker->thread,	NULL, wf_workerThread, (void *)worker);

	// and register the worker
	wf.workersCount++;
	worker->id = wf.workersCount;
	
	if(wf.firstWorker == NULL) {
		wf.firstWorker = worker;
		wf.lastWorker = worker;
		worker->prev = NULL;
		worker->next = NULL;
	} else {
		worker->prev = wf.lastWorker;
		wf.lastWorker->next = worker;
		wf.lastWorker = worker;
	}
}

/*
*	Bootstrap factory: create workers and register it
*/
void wf_bootstrap() {
	int workers = DEFAULT_WORKERS;
	
	wf.firstWorker = NULL;
	wf.lastWorker = NULL;
	wf.workersCount = 0;
	
	while(workers-- > 0) {
		wf_workerHotCreate();
	}
}